# src/automation/__init__.py
from .drission_bot import DrissionBot
from .get import get_bot
from .selenium_bot import SeleniumBot

# only import __all__ when using from automation import *
__all__ = ["SeleniumBot", "DrissionBot", "get_bot"]



# v2dl/__init__.py
from .config import Config, ConfigManager, RuntimeConfig
from .error import DownloadError, FileProcessingError, ScrapeError
from .logger import setup_logging
from .scrapper import ScrapeHandler
from .utils.utils import ThreadingService
from .v2dl import ScrapeManager
from .version import __version__
from .web_bot import get_bot

__all__ = [
    "Config",
    "RuntimeConfig",
    "ConfigManager",
    "setup_logging",
    "ThreadingService",
    "ScrapeHandler",
    "ScrapeManager",
    "ScrapeError",
    "FileProcessingError",
    "DownloadError",
    "get_bot",
    "__version__",
]



import logging
import os
import re
import sys
import threading
import time
from dataclasses import dataclass
from pathlib import Path
from queue import Queue
from typing import Any, Callable, Generic, Optional, TypeVar, Union
from urllib.parse import parse_qs, urlencode, urlparse, urlunparse

import requests
from lxml import html


class LinkParser:
    """Tool class parses URL."""

    @staticmethod
    def parse_input_url(url: str) -> tuple[list[str], int]:
        parsed_url = urlparse(url)
        path_parts = parsed_url.path.split("/")
        query_params = parse_qs(parsed_url.query)
        start_page: int = int(query_params.get("page", [1])[0])  # default page=1
        return path_parts, start_page

    @staticmethod
    def parse_html(html_content: str, logger: logging.Logger) -> html.HtmlElement | None:
        if "Failed" in html_content:
            return None

        try:
            return html.fromstring(html_content)
        except Exception as e:
            logger.error("Error parsing HTML content: %s", e)
            return None

    @staticmethod
    def get_max_page(tree: html.HtmlElement) -> int:
        """Parse pagination count."""
        page_links = tree.xpath(
            '//li[@class="page-item"]/a[@class="page-link" and string-length(text()) <= 2]/@href'
        )

        if not page_links:
            return 1

        page_numbers = []
        for link in page_links:
            match = re.search(r"page=(\d+)", link)
            if match:
                page_number = int(match.group(1))
            else:
                page_number = 1
            page_numbers.append(page_number)

        return max(page_numbers)

    @staticmethod
    def add_page_num(url: str, page: int) -> str:
        parsed_url = urlparse(url)  # 解析 URL
        query_params = parse_qs(parsed_url.query)  # 解析查詢參數
        query_params["page"] = [str(page)]  # 修改頁碼

        new_query = urlencode(query_params, doseq=True)  # 組合成字串
        new_url = parsed_url._replace(query=new_query)  # 替換頁碼

        # Example
        # url = "https://example.com/search?q=test&sort=asc", page = 3
        # parsed_url: ParseResult(scheme='https', netloc='example.com', path='/search', params='', query='q=test&sort=asc', fragment='')
        # query_params: {'q': ['test'], 'sort': ['asc'], 'page': ['3']}
        # new_query: 'q=test&sort=asc&page=3'
        # new_url: ParseResult(scheme='https', netloc='example.com', path='/search', params='', query='q=test&sort=asc&page=3', fragment='')
        # urlunparse: 'https://example.com/search?q=test&sort=asc&page=3'
        return urlunparse(new_url)

    @staticmethod
    def remove_page_num(url: str) -> str:
        """remove ?page=d or &page=d from URL."""
        # Parse the URL
        parsed_url = urlparse(url)
        query_params = parse_qs(parsed_url.query)

        # Remove the 'page' parameter if it exists
        if "page" in query_params:
            del query_params["page"]

        # Rebuild the query string without 'page'
        new_query = urlencode(query_params, doseq=True)

        # Rebuild the full URL
        new_url = urlunparse(parsed_url._replace(query=new_query))
        return new_url


# Generic type for task input and task result
TI = TypeVar("TI", bound=Any)
TR = TypeVar("TR", bound=Any)


@dataclass
class ThreadJob(Generic[TI]):
    """Generic task container."""

    task_id: str
    params: TI
    job: Union[Callable[..., TR], Any] = None  # type: ignore


class ThreadingService(Generic[TI, TR]):
    """Generic service for processing tasks with multiple workers."""

    def __init__(self, logger: logging.Logger, num_workers: int = 1):
        self.task_queue: Queue[Optional[ThreadJob[TI]]] = Queue()
        self.logger = logger
        self.num_workers = num_workers
        self.worker_threads: list[threading.Thread] = []
        self.results: dict[str, TR] = {}
        self._lock = threading.Lock()

    def start_workers(self):
        """Start up multiple worker threads to listen for tasks."""
        for _ in range(self.num_workers):
            worker = threading.Thread(target=self._task_worker, daemon=True)
            self.worker_threads.append(worker)
            worker.start()

    def _task_worker(self):
        """Worker function to process tasks from the queue."""
        while True:
            task = self.task_queue.get()
            if task is None:
                break  # exit signal received

            try:
                if task.job:
                    result: Any = task.job(task.task_id, task.params, self.logger)
                    with self._lock:
                        self.results[task.task_id] = result
            except Exception as e:
                self.logger.error("Error processing task %s: %s", task.task_id, e)
            finally:
                self.task_queue.task_done()

    def add_task(self, task_id: str, params: TI, job: Any) -> None:
        """Add task to queue with specific parameters and job."""
        task = ThreadJob(task_id=task_id, params=params, job=job)
        self.task_queue.put(task)

    def get_result(self, task_id: str) -> Optional[TR]:
        """Get the result of a specific task."""
        with self._lock:
            return self.results.get(task_id)

    def wait_completion(self):
        """Block until all tasks are done and stop all workers."""
        self.task_queue.join()

        # Signal all workers to exit
        for _ in range(self.num_workers):
            self.task_queue.put(None)

        # Wait for all worker threads to finish
        for worker in self.worker_threads:
            worker.join()


class AlbumTracker:
    """Download log in units of albums."""

    def __init__(self, download_log: str):
        self.album_log_path = download_log

    def is_downloaded(self, album_url: str) -> bool:
        if os.path.exists(self.album_log_path):
            with open(self.album_log_path) as f:
                downloaded_albums = f.read().splitlines()
            return album_url in downloaded_albums
        return False

    def log_downloaded(self, album_url: str):
        album_url = LinkParser.remove_page_num(album_url)
        if not self.is_downloaded(album_url):
            with open(self.album_log_path, "a") as f:
                f.write(album_url + "\n")


def threading_download_job(task_id: str, params: tuple, logger: logging.Logger) -> bool:
    """Download single photo job for threading service.

    Args:
        task_id (str): Task identifier
        params (tuple): (url, alt, destination, headers, rate_limit, no_skip)
                       destination is the parent directory for saving files
        logger (logging.Logger): Logger instance

    Returns:
        bool: True if download successful, False otherwise
    """
    try:
        url, alt, destination, headers, rate_limit, no_skip = params

        # obtain album_name from task_id (the format of task_id is "album_name_index")
        album_name = task_id.rsplit("_", 1)[0]

        # setup paths
        folder = destination / Path(album_name)
        folder.mkdir(parents=True, exist_ok=True)
        filename = re.sub(r'[<>:"/\\|?*]', "", alt)  # Remove invalid characters
        file_path = folder / f"{filename}.{get_image_extension(url)}"

        if file_path.exists() and not no_skip:
            logger.info("File already exists: '%s'", file_path)
            return True

        return download_image(url, file_path, headers, rate_limit, logger)

    except Exception as e:
        logger.error("Error downloading photo %s: %s", task_id, e)
        return False


def threading_process_data_job(task_id: str, params: dict, logger: logging.Logger) -> dict:
    # Process data and return results
    return {"processed": True, "data": params}


def download_album(  # noqa: PLR0913
    album_name: str,
    image_links: list[tuple[str, str]],
    destination: str,
    headers: dict,
    rate_limit: int,
    no_skip: bool,
    logger: logging.Logger,
):
    """Download images from image links.

    Save images to a folder named after the album, existing files would be skipped.

    Args:
        album_name (str): Name of album folder.
        image_links (list[tuple[str, str]]): List of tuples with image URLs and corresponding alt text for filenames.
        destination (str): Download parent directory of album folder.
        headers (dict): Download request headers.
        rate_limit (int): Download rate limits.
        no_skip (bool): Do not skip downloaded files.
        logger (logging.Logger): Logger.
    """
    folder = destination / Path(album_name)
    folder.mkdir(parents=True, exist_ok=True)

    for url, alt in image_links:
        filename = re.sub(r'[<>:"/\\|?*]', "", alt)  # Remove invalid characters
        file_path = folder / f"{filename}.{get_image_extension(url)}"

        if file_path.exists() and not no_skip:
            logger.info("File already exists: '%s'", file_path)
            continue

        # requests module will log download url
        if download_image(url, file_path, headers, rate_limit, logger):
            pass


def download_image(
    url: str,
    save_path: Path,
    headers: dict,
    rate_limit: int,
    logger: logging.Logger,
) -> bool:
    """Error control subfunction for download files.

    Return `True` for successful download, else `False`.
    """
    try:
        download(url, save_path, headers, rate_limit)
        logger.info("Downloaded: '%s'", save_path)
        return True
    except requests.exceptions.HTTPError as http_err:
        logger.error("HTTP error occurred: %s", http_err)
        return False
    except Exception as e:
        logger.error("An error occurred while downloading url '%s': %s", url, e)
        return False


def download(url: str, save_path: Path, headers: dict, speed_limit_kbps: int = 1536) -> None:
    """Download with speed limit function.

    Default speed limit is 1536 KBps (1.5 MBps).
    """

    if headers is None:
        headers = {}
    chunk_size = 1024
    speed_limit_bps = speed_limit_kbps * 1024  # 轉換為 bytes per second

    response = requests.get(url, stream=True, headers=headers)
    response.raise_for_status()  # 確認請求成功

    with open(save_path, "wb") as file:
        start_time = time.time()
        downloaded = 0

        for chunk in response.iter_content(chunk_size=chunk_size):
            file.write(chunk)
            downloaded += len(chunk)

            elapsed_time = time.time() - start_time
            expected_time = downloaded / speed_limit_bps

            if elapsed_time < expected_time:
                time.sleep(expected_time - elapsed_time)


def get_image_extension(url: str) -> str:
    """Get the extension of url.

    If there is not an extension, return default value "jpg".
    """
    image_extensions = r"(?:[^.]|^)\.(jpg|jpeg|png|gif|bmp|webp|tiff|svg)$"

    match = re.search(image_extensions, url, re.IGNORECASE)

    if match:
        return match.group(1)
    else:
        # 如果沒找到，返回預設值
        return "jpg"


def check_input_file(input_path: str):
    if input_path and not os.path.isfile(input_path):
        logging.error("Input file %s does not exist.", input_path)
        sys.exit(1)
    else:
        logging.info("Input file %s exists and is accessible.", input_path)



import os
import platform
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import yaml
from dotenv import load_dotenv


@dataclass
class RuntimeConfig:
    url: str
    input_file: str
    bot_type: str
    chrome_args: list[str] | None
    user_agent: str | None
    terminate: bool
    download_service: Any
    dry_run: bool
    logger: Any
    log_level: int
    no_skip: bool = False
    use_chrome_default_profile: bool = False


@dataclass
class DownloadConfig:
    min_scroll_length: int
    max_scroll_length: int
    min_scroll_step: int
    max_scroll_step: int
    rate_limit: int
    download_dir: str


@dataclass
class PathConfig:
    download_log: str
    system_log: str


@dataclass
class ChromeConfig:
    exec_path: str
    profile_path: str


@dataclass
class Config:
    download: DownloadConfig
    paths: PathConfig
    chrome: ChromeConfig


class ConfigManager:
    """Load and process configs based on user platform.

    The DEFAULT_CONFIG is a nested dict, after processing, the ConfigManager.load() returns a
    Config dataclass consists of DownloadConfig, PathConfig, ChromeConfig dataclasses.
    """

    def __init__(self, config: dict[str, dict[str, Any]], config_dir: str | None = None):
        self.config = config
        self.config_dir = config_dir

    def load(self) -> Config:
        """Load configuration from files and environment."""
        system_config_dir = ConfigManager.get_system_config_dir()
        if self.config_dir is not None:  # overwrite the config_dir
            system_config_dir = Path(self.config_dir)
        system_config_dir.mkdir(parents=True, exist_ok=True)

        custom_config_path = system_config_dir / "config.yaml"
        custom_env_path = system_config_dir / ".env"

        # Load environment variables
        if custom_env_path.exists():
            load_dotenv(custom_env_path)

        # Load and merge configurations
        if custom_config_path.exists():
            with open(custom_config_path) as f:
                custom_config = yaml.safe_load(f)
                if custom_config:  # not empty
                    self.config = ConfigManager._merge_config(self.config, custom_config)

        # Check file paths
        for key, path in self.config["paths"].items():
            self.config["paths"][key] = self.resolve_path(path, system_config_dir)

        self.config["chrome"]["profile_path"] = self.resolve_path(
            self.config["chrome"]["profile_path"], system_config_dir
        )

        # Check download_dir path
        download_dir = self.config["download"].get("download_dir", "").strip()
        self.config["download"]["download_dir"] = self._get_download_dir(download_dir)

        return Config(
            download=DownloadConfig(**self.config["download"]),
            paths=PathConfig(**self.config["paths"]),
            chrome=ChromeConfig(
                exec_path=ConfigManager._get_chrome_exec_path(self.config),
                profile_path=self.config["chrome"]["profile_path"],
            ),
        )

    def resolve_path(self, path, base_dir):
        """Resolve '~', add path with base_dir if input is not absolute path."""
        path = os.path.expanduser(path)
        return os.path.join(base_dir, path) if not os.path.isabs(path) else path

    @staticmethod
    def get_system_config_dir() -> Path:
        """Return the config directory."""
        if platform.system() == "Windows":
            base = os.getenv("APPDATA", "")
        else:
            base = os.path.expanduser("~/.config")
        return Path(base) / "v2dl"

    @staticmethod
    def get_default_download_dir() -> Path:
        return Path.home() / "Downloads"

    def _get_download_dir(self, download_dir: str) -> str:
        sys_dl_dir = ConfigManager.get_default_download_dir()
        result_dir = self.resolve_path(download_dir, sys_dl_dir) if download_dir else sys_dl_dir
        result_dir = Path(result_dir)
        result_dir.mkdir(parents=True, exist_ok=True)
        return str(result_dir)

    @staticmethod
    def _get_chrome_exec_path(config_data: dict) -> str:
        current_os = platform.system()
        exec_path = config_data["chrome"]["exec_path"].get(current_os)
        if not exec_path:
            raise ValueError(f"Unsupported OS: {current_os}")
        return exec_path

    @staticmethod
    def _merge_config(base: dict[str, Any], custom: dict[str, Any]) -> dict:
        """Recursively merge custom config into base config."""
        for key, value in custom.items():
            if isinstance(value, dict) and key in base:
                ConfigManager._merge_config(base[key], value)
            else:
                base[key] = value
        return base



import os
import random
import time
from abc import ABC, abstractmethod
from typing import Any

from ..config import Config, RuntimeConfig
from ..utils.security_utils import AccountManager, Encryptor


class BaseBot(ABC):
    """Abstract base class for bots, defining shared behaviors."""

    def __init__(self, runtime_config: RuntimeConfig, base_config: Config):
        self.runtime_config = runtime_config
        self.config = base_config
        self.close_browser = runtime_config.terminate
        self.logger = runtime_config.logger

        self.encryptor = Encryptor(self.logger)
        self.account_manager = AccountManager(self.encryptor, self.logger)
        self.private_key, self.public_key = self.encryptor.load_keys()

        self.new_profile = False

    @abstractmethod
    def init_driver(self) -> Any:
        """Initialize the browser driver."""

    @abstractmethod
    def close_driver(self):
        """Close the browser and handle cleanup."""

    def prepare_chrome_profile(self):
        user_data_dir = self.config.chrome.profile_path

        if not os.path.exists(user_data_dir):
            os.makedirs(user_data_dir)
            self.new_profile = True
        else:
            self.new_profile = False

        return user_data_dir

    def auto_page_scroll(
        self, url: str, max_retry: int = 3, page_sleep: int = 5, fast_scroll: bool = True
    ) -> str:
        """Request handling with retries. To be implemented in subclasses."""
        raise NotImplementedError("Subclasses must implement automated retry logic.")

    def handle_login(self):
        """Login logic, implemented by subclasses."""
        raise NotImplementedError("Subclasses must implement login logic.")

    def human_like_type(self, element, text):
        """Simulate human-like typing into a field."""
        raise NotImplementedError("Subclasses must implement scroll behavior.")

    def scroll_page(self):
        """Simulate human-like scrolling behavior."""
        raise NotImplementedError("Subclasses must implement scroll behavior.")


class BaseBehavior:
    pause_time = (0.1, 0.3)

    @staticmethod
    def random_sleep(min_time=1.0, max_time=5.0):
        time.sleep(random.uniform(min_time, max_time))


class BaseScroll:
    def __init__(self, config, logger):
        self.config = config
        self.logger = logger
        self.scroll_position = 0
        self.last_content_height = 0
        self.successive_scroll_count = 0
        self.max_successive_scrolls = random.randint(5, 10)



import time

from .drission_bot import DrissionBot
from .selenium_bot import SeleniumBot


def get_bot(runtime_config, config):
    bot_classes = {"selenium": SeleniumBot, "drission": DrissionBot}

    bot_type = runtime_config.bot_type
    close_browser = runtime_config.terminate
    logger = runtime_config.logger

    if bot_type not in bot_classes:
        raise ValueError(f"Unsupported automator type: {bot_type}")

    bot = bot_classes[bot_type](runtime_config, config)

    if bot.new_profile:
        init_new_profile(bot)
    return bot


def init_new_profile(bot):
    # visit some websites for new chrome profile
    websites = [
        "https://www.google.com",
        "https://www.youtube.com",
        "https://www.wikipedia.org",
    ]

    for url in websites:
        if isinstance(bot, DrissionBot):
            bot.page.get(url)
        elif isinstance(bot, SeleniumBot):
            bot.driver.get(url)

        time.sleep(4)



# Do NOT import utils/file_utils
import logging
import os
from typing import ClassVar

from colorama import Fore, Style, init

init()


class CustomFormatter(logging.Formatter):
    COLORS: ClassVar[dict] = {
        logging.DEBUG: Fore.LIGHTBLACK_EX,
        logging.INFO: Fore.WHITE,
        logging.WARNING: Fore.YELLOW,
        logging.ERROR: Fore.RED,
        logging.CRITICAL: Fore.RED + Style.BRIGHT,
    }
    GREEN = Fore.GREEN
    RESET = Style.RESET_ALL

    def __init__(self, use_color=True):
        super().__init__()
        self.use_color = use_color

    def format(self, record):
        if self.use_color:
            color = self.COLORS.get(record.levelno, self.RESET)
            levelname = record.levelname.lower()
            return (
                f"[{self.GREEN}{self.formatTime(record, '%H:%M:%S')}{self.RESET}]"
                f"[{color}{levelname}{self.RESET}] - {record.getMessage()}"
            )
        else:
            # Convert levelname to lowercase for file logs
            levelname = record.levelname.lower()
            return f"[{self.formatTime(record, '%H:%M:%S')}][{levelname}] - {record.getMessage()}"


def setup_logging(level, log_path, no_archive=False):
    """level: [logging.LOGLEVEL]
    args: [bool], no_archive..
    """
    # Clear any existing handlers
    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)

    # Create formatters
    color_formatter = CustomFormatter(use_color=True)
    plain_formatter = CustomFormatter(use_color=False)

    # Console handler
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(color_formatter)
    logging.getLogger().addHandler(console_handler)

    # File handler
    if not no_archive:
        log_dir = os.path.dirname(log_path)
        os.makedirs(log_dir, exist_ok=True)
        file_handler = logging.FileHandler(log_path, encoding="utf-8")
        file_handler.setFormatter(plain_formatter)
        logging.getLogger().addHandler(file_handler)

    # Set log level
    logging.getLogger().setLevel(level)


if __name__ == "__main__":
    # Set up logging
    setup_logging(logging.DEBUG, "test_logger.log")

    # Create a logger
    logger = logging.getLogger(__name__)

    # Log messages to test the configuration
    logger.debug("Debug message")
    logger.info("This is an info message.")
    logger.warning("Warning message")
    logger.error("This is an error message.")
    logger.critical("Critical message")



import re
import time
from abc import ABC, abstractmethod
from typing import ClassVar, Generic, Literal, TypeAlias, TypeVar, Union, overload

from lxml import html

from .config import Config, RuntimeConfig
from .const import BASE_URL, HEADERS
from .utils.utils import AlbumTracker, LinkParser, threading_download_job

# Manage return types of each scraper here
AlbumLink: TypeAlias = str
ImageLinkAndALT: TypeAlias = tuple[str, str]
LinkType = TypeVar("LinkType", AlbumLink, ImageLinkAndALT)

# Define literal types for scraping types
ScrapeType = Literal["album_list", "album_image"]


class ScrapeHandler:
    """Handles all scraper behaviors."""

    # Defines the mapping from url part to scrape method.
    URL_HANDLERS: ClassVar[dict[str, ScrapeType]] = {
        "album": "album_image",
        "actor": "album_list",
        "company": "album_list",
        "category": "album_list",
        "country": "album_list",
    }

    def __init__(self, runtime_config: RuntimeConfig, base_config: Config, web_bot):
        self.web_bot = web_bot
        self.logger = runtime_config.logger
        self.runtime_config = runtime_config
        self.strategies: dict[ScrapeType, BaseScraper] = {
            "album_list": AlbumScraper(runtime_config, base_config, web_bot),
            "album_image": ImageScraper(runtime_config, base_config, web_bot),
        }

        self.album_tracker = AlbumTracker(base_config.paths.download_log)
        self.path_parts, self.start_page = LinkParser.parse_input_url(runtime_config.url)

    def scrape(self, url: str, dry_run: bool = False) -> None:
        """Main entry point for scraping operations."""
        scrape_type = self._get_scrape_type()
        if scrape_type == "album_list":
            self.scrape_album_list(url, self.start_page, dry_run)
        else:
            self.scrape_album(url, self.start_page, dry_run)

    def scrape_album_list(self, url: str, start_page: int, dry_run: bool) -> None:
        """Handle scraping of album lists."""
        album_links = self._real_scrape(url, start_page, "album_list")
        self.logger.info("Found %d albums", len(album_links))

        for album_url in album_links:
            if dry_run:
                self.logger.info("[DRY RUN] Album URL: %s", album_url)
            else:
                self.scrape_album(album_url, 1, dry_run)

    def scrape_album(self, album_url: str, start_page: int, dry_run: bool) -> None:
        """Handle scraping of a single album page."""
        if self.album_tracker.is_downloaded(album_url) and not self.runtime_config.no_skip:
            self.logger.info("Album %s already downloaded, skipping.", album_url)
            return

        image_links = self._real_scrape(album_url, start_page, "album_image")
        if not image_links:
            return

        album_name = re.sub(r"\s*\d+$", "", image_links[0][1])
        self.logger.info("Found %d images in album %s", len(image_links), album_name)

        if dry_run:
            for link, alt in image_links:
                self.logger.info("[DRY RUN] Image URL: %s", link)
        else:
            self.album_tracker.log_downloaded(album_url)

    @overload
    def _real_scrape(
        self, url: str, start_page: int, scrape_type: Literal["album_list"], **kwargs
    ) -> list[AlbumLink]: ...

    @overload
    def _real_scrape(
        self, url: str, start_page: int, scrape_type: Literal["album_image"], **kwargs
    ) -> list[ImageLinkAndALT]: ...

    def _real_scrape(
        self,
        url: str,
        start_page: int,
        scrape_type: ScrapeType,
        **kwargs,
    ) -> list[AlbumLink] | list[ImageLinkAndALT]:
        """Scrape pages for links using the appropriate strategy."""
        strategy = self.strategies[scrape_type]
        self.logger.info(
            "Starting to scrape %s links from %s", "album" if scrape_type else "image", url
        )

        page_result: Union[list[AlbumLink], list[ImageLinkAndALT]] = []
        page = start_page

        while True:
            full_url = LinkParser.add_page_num(url, page)
            html_content = self.web_bot.auto_page_scroll(full_url, page_sleep=0)
            tree = LinkParser.parse_html(html_content, self.logger)

            if tree is None:
                break

            # log entering a page
            self.logger.info("Fetching content from %s", full_url)
            page_links = tree.xpath(strategy.get_xpath())

            # log no images
            if not page_links:
                self.logger.info(
                    "No more %s found on page %d",
                    "albums" if scrape_type == "album_list" else "images",
                    page,
                )
                break

            strategy.process_page_links(page_links, page_result, tree, page)

            if page >= LinkParser.get_max_page(tree):
                self.logger.info("Reach last page, stopping")
                break

            page = self._handle_pagination(page, **kwargs)

        return page_result

    def _handle_pagination(
        self,
        current_page: int,
        max_consecutive_page: int = 3,
        consecutive_sleep: int = 15,
    ) -> int:
        """Handle pagination logic including sleep for consecutive pages."""
        next_page = current_page + 1
        if next_page % max_consecutive_page == 0:
            time.sleep(consecutive_sleep)
        return next_page

    def _get_scrape_type(self):
        """Get the appropriate handler method based on URL path."""
        for part in self.path_parts:
            if part in self.URL_HANDLERS:
                return self.URL_HANDLERS[part]
        raise ValueError(f"Unsupported URL type: {self.runtime_config.url}")


class BaseScraper(Generic[LinkType], ABC):
    """Abstract base class for different scraping strategies."""

    def __init__(self, runtime_config: RuntimeConfig, base_config: Config, web_bot):
        self.runtime_config = runtime_config
        self.config = base_config
        self.web_bot = web_bot
        self.download_service = runtime_config.download_service
        self.logger = runtime_config.logger

    @abstractmethod
    def get_xpath(self) -> str:
        """Return xpath for the specific strategy."""

    @abstractmethod
    def process_page_links(
        self,
        page_links: list[str],
        page_result: list[LinkType],
        tree: html.HtmlElement,
        page: int,
        **kwargs,
    ) -> None:
        """Process links found on the page."""


class AlbumScraper(BaseScraper[AlbumLink]):
    """Strategy for scraping album list pages."""

    XPATH_ALBUM_LIST = '//a[@class="media-cover"]/@href'

    def get_xpath(self) -> str:
        return self.XPATH_ALBUM_LIST

    def process_page_links(
        self,
        page_links: list[str],
        page_result: list[AlbumLink],
        tree: html.HtmlElement,
        page: int,
        **kwargs,
    ) -> None:
        page_result.extend([BASE_URL + album_link for album_link in page_links])
        self.logger.info("Found %d albums on page %d", len(page_links), page)


class ImageScraper(BaseScraper[ImageLinkAndALT]):
    """Strategy for scraping album image pages."""

    XPATH_ALBUM = '//div[@class="album-photo my-2"]/img/@data-src'
    XPATH_ALTS = '//div[@class="album-photo my-2"]/img/@alt'

    def __init__(self, runtime_config: RuntimeConfig, base_config: Config, web_bot):
        super().__init__(runtime_config, base_config, web_bot)
        self.dry_run = runtime_config.dry_run
        self.alt_counter = 0

    def get_xpath(self) -> str:
        return self.XPATH_ALBUM

    def process_page_links(
        self,
        page_links: list[str],
        page_result: list[ImageLinkAndALT],
        tree: html.HtmlElement,
        page: int,
        **kwargs,
    ) -> None:
        alts: list[str] = tree.xpath(self.XPATH_ALTS)

        # Handle missing alt texts
        if len(alts) < len(page_links):
            missing_alts = [str(i + self.alt_counter) for i in range(len(page_links) - len(alts))]
            alts.extend(missing_alts)
            self.alt_counter += len(missing_alts)

        page_result.extend(zip(page_links, alts))

        # Handle downloads if not in dry run mode
        if not self.dry_run:
            album_name = self._extract_album_name(alts)

            # assign download job for each image
            for i, (url, alt) in enumerate(zip(page_links, alts)):
                self.download_service.add_task(
                    task_id=f"{album_name}_{i}",
                    params=(
                        url,
                        alt,
                        self.config.download.download_dir,
                        HEADERS,
                        self.config.download.rate_limit,
                        self.runtime_config.no_skip,
                    ),
                    job=threading_download_job,
                )

        self.logger.info("Found %d images on page %d", len(page_links), page)

    @staticmethod
    def _extract_album_name(alts: list[str]) -> str:
        album_name = next((alt for alt in alts if not alt.isdigit()), None)
        if album_name:
            album_name = re.sub(r"\s*\d*$", "", album_name).strip()
        if not album_name:
            album_name = BASE_URL.rstrip("/").split("/")[-1]
        return album_name



import logging
import sys

from .account_cli import cli
from .config import Config, ConfigManager, RuntimeConfig
from .const import DEFAULT_CONFIG
from .error import ScrapeError
from .logger import setup_logging
from .option import parse_arguments
from .scrapper import ScrapeHandler
from .utils.utils import ThreadingService, check_input_file
from .version import __version__
from .web_bot import get_bot


class ScrapeManager:
    """Manage the starting and ending of the scraper."""

    def __init__(self, runtime_config: RuntimeConfig, base_config: Config, web_bot):
        self.runtime_config = runtime_config
        self.base_config = base_config

        self.web_bot = web_bot
        self.dry_run = runtime_config.dry_run
        self.logger = runtime_config.logger

        # 初始化
        self.download_service: ThreadingService = runtime_config.download_service

        if not self.dry_run:
            self.download_service.start_workers()

    def start_scraping(self):
        """Start scraping based on URL type."""
        try:
            urls = self._load_urls()
            for url in urls:
                self.runtime_config.url = url
                link_scraper = ScrapeHandler(self.runtime_config, self.base_config, self.web_bot)
                link_scraper.scrape(url, self.dry_run)
        except ScrapeError as e:
            self.logger.exception("Scraping error: '%s'", e)
        finally:
            if not self.dry_run:
                self.download_service.wait_completion()
            self.web_bot.close_driver()

    def _load_urls(self):
        """Load URLs from runtime_config (URL or txt file)."""
        if self.runtime_config.input_file:
            with open(self.runtime_config.input_file) as file:
                urls = [line.strip() for line in file if line.strip()]
        else:
            urls = [self.runtime_config.url]
        return urls


def main():
    args, log_level = parse_arguments()

    if args.version:
        print(f"{__version__}")
        sys.exit(0)

    if args.input_file:
        check_input_file(args.input_file)

    if args.account:
        cli()

    app_config = ConfigManager(DEFAULT_CONFIG).load()

    setup_logging(log_level, log_path=app_config.paths.system_log)
    logger = logging.getLogger(__name__)
    download_service: ThreadingService = ThreadingService(logger, num_workers=3)

    runtime_config = RuntimeConfig(
        url=args.url,
        input_file=args.input_file,
        bot_type=args.bot_type,
        chrome_args=args.chrome_args,
        user_agent=args.user_agent,
        use_chrome_default_profile=args.use_default_chrome_profile,
        terminate=args.terminate,
        download_service=download_service,
        dry_run=args.dry_run,
        logger=logger,
        log_level=log_level,
        no_skip=args.no_skip,
    )

    web_bot = get_bot(runtime_config, app_config)
    scraper = ScrapeManager(runtime_config, app_config, web_bot)
    scraper.start_scraping()



import atexit
import base64
import os
import random
import secrets
import sys
import threading
from datetime import datetime, timedelta

import yaml
from dotenv import load_dotenv, set_key
from nacl.public import PrivateKey, PublicKey, SealedBox
from nacl.pwhash import argon2id
from nacl.secret import SecretBox
from nacl.utils import random as nacl_random

from ..config import ConfigManager
from ..error import SecurityError


class Encryptor:
    KEY_BYTES = 32
    SALT_BYTES = 16
    NONCE_BYTES = 24
    KDF_OPS_LIMIT = 2**10
    KDF_MEM_LIMIT = 2**16

    def __init__(self, logger) -> None:
        self.logger = logger
        self.config = self._initialize_config()
        self.file_handler = SecureFileHandler()
        self.custom_env_path = os.path.join(ConfigManager.get_system_config_dir(), ".env")
        self.__ensure_secure_folder()
        self._generate_keypair()

    def _initialize_config(self) -> dict:
        """Initialize configuration settings."""
        base_dir = ConfigManager.get_system_config_dir()
        self.logger.debug("Initializing config with base directory: %s", base_dir)
        return {
            "key_folder": os.path.join(base_dir, ".keys"),
            "env_path": os.path.join(base_dir, ".env"),
            "master_key_file": os.path.join(base_dir, ".keys", "master_key.enc"),
            "private_key_file": os.path.join(base_dir, ".keys", "private_key.pem"),
            "public_key_file": os.path.join(base_dir, ".keys", "public_key.pem"),
        }

    def encrypt_master_key(self, master_key: bytes) -> tuple[bytes, bytes, bytes]:
        """Encrypt the master key using scrypt."""
        salt = secrets.token_bytes(self.SALT_BYTES)
        encryption_key = secrets.token_bytes(self.KEY_BYTES)

        # Derive the encryption key using scrypt
        import time

        t = time.time()
        derived_key = argon2id.kdf(
            self.KEY_BYTES,
            encryption_key,
            salt,
            opslimit=self.KDF_OPS_LIMIT,
            memlimit=self.KDF_MEM_LIMIT,
        )
        print(time.time() - t)
        self.logger.debug("Derived encryption key with salt: %s", salt.hex())
        box = SecretBox(derived_key)
        nonce = nacl_random(self.NONCE_BYTES)
        encrypted_master_key = box.encrypt(master_key, nonce)

        # Clean sensitive data
        derived_key = bytearray(len(derived_key))
        self.logger.info("Master key encryption successful")

        return encrypted_master_key, salt, encryption_key

    def decrypt_master_key(
        self, encrypted_master_key: bytes, salt: bytes, encryption_key: bytes
    ) -> bytes:
        """Decrypt the master key using scrypt."""
        # Derive the decryption key using scrypt
        derived_key = argon2id.kdf(
            self.KEY_BYTES,
            encryption_key,
            salt,
            opslimit=self.KDF_OPS_LIMIT,
            memlimit=self.KDF_MEM_LIMIT,
        )

        box = SecretBox(derived_key)
        master_key = box.decrypt(encrypted_master_key)

        # Clean sensitive data
        derived_key = bytearray(len(derived_key))
        self.logger.info("Master key decryption successful")

        return master_key

    def encrypt_password(self, password: str, public_key: PublicKey) -> str:
        """Encrypt a password using the public key."""
        sealed_box = SealedBox(public_key)
        encrypted = sealed_box.encrypt(password.encode())
        self.logger.info("Password encryption successful")

        return base64.b64encode(encrypted).decode("utf-8")

    def decrypt_password(self, encrypted_password: str, private_key: PrivateKey) -> str:
        """Decrypt a password using the private key."""
        encrypted = base64.b64decode(encrypted_password)
        sealed_box = SealedBox(private_key)
        decrypted = sealed_box.decrypt(encrypted)
        return decrypted.decode()

    def _generate_keypair(self) -> None:
        """Generate and store a new keypair with the master key."""
        if os.path.exists(self.config["private_key_file"]) and os.path.exists(
            self.config["public_key_file"]
        ):
            self.logger.info("Key pair already exists.")
            return

        try:
            self.__ensure_secure_folder()

            # Generate keys
            private_key = PrivateKey.generate()
            public_key = private_key.public_key
            master_key = secrets.token_bytes(self.KEY_BYTES)

            # Encrypt master key and private key
            encrypted_master_key, salt, encryption_key = self.encrypt_master_key(master_key)
            encrypted_private_key = self._encrypt_private_key(private_key, master_key)

            # Store keys
            self._store_keys(encrypted_master_key, encrypted_private_key, public_key)
            self._store_encryption_params(salt, encryption_key)

            # Clean sensitive data
            self._secure_cleanup([master_key, encryption_key])
            self.logger.info("Key pair has been successfully generated and stored.")

        except Exception as e:
            self.logger.error("Key generation failed: %s", str(e))
            raise SecurityError("Key generation failed") from e

    def _encrypt_private_key(self, private_key: PrivateKey, master_key: bytes) -> bytes:
        """Encrypt the private key using the master key."""
        box = SecretBox(master_key)
        nonce = nacl_random(self.NONCE_BYTES)
        return box.encrypt(private_key.encode(), nonce)

    def _store_keys(
        self, encrypted_master_key: bytes, encrypted_private_key: bytes, public_key: PublicKey
    ) -> None:
        """Store all keys securely."""
        self.logger.debug("Storing keys securely.")
        self.file_handler.write_secure_file(self.config["master_key_file"], encrypted_master_key)
        self.file_handler.write_secure_file(self.config["private_key_file"], encrypted_private_key)
        self.file_handler.write_secure_file(
            self.config["public_key_file"], public_key.encode(), 0o644
        )

    def _store_encryption_params(self, salt: bytes, encryption_key: bytes) -> None:
        """Store encryption parameters in the environment file."""
        self.logger.debug("Storing encryption parameters.")
        load_dotenv(self.config["env_path"])
        salt_b64 = base64.b64encode(salt).decode("utf-8")
        enc_key_b64 = base64.b64encode(encryption_key).decode("utf-8")

        set_key(self.config["env_path"], "SALT", salt_b64)
        set_key(self.config["env_path"], "ENCRYPTION_KEY", enc_key_b64)

    @staticmethod
    def _secure_cleanup(sensitive_data: list[bytes]) -> None:
        """Securely clear sensitive data from memory."""
        for data in sensitive_data:
            data = bytearray(len(data))  # noqa: PLW2901

    def load_keys(self) -> tuple[PrivateKey, PublicKey]:
        """Load and validate the keypair."""
        try:
            # Load encrypted keys
            self.logger.debug("Loading and validating keys.")
            encrypted_master_key = self.file_handler.read_secure_file(
                self.config["master_key_file"]
            )
            encrypted_private_key = self.file_handler.read_secure_file(
                self.config["private_key_file"]
            )
            public_key_bytes = self.file_handler.read_secure_file(self.config["public_key_file"])

            # Decrypt keys
            master_key = self._decrypt_master_key(encrypted_master_key)
            private_key = self._decrypt_private_key(encrypted_private_key, master_key)
            public_key = PublicKey(public_key_bytes)

            # Validate keypair
            self._validate_keypair(private_key, public_key)

            # Clean sensitive data
            self._secure_cleanup([master_key])
            self.logger.info("Keys loaded and validated successfully.")

            return private_key, public_key

        except Exception as e:
            self.logger.error("Key loading failed: %s", str(e))
            raise SecurityError("Key loading failed") from e

    def load_and_validate_env(self) -> tuple[str, str]:
        load_dotenv(self.config["env_path"])

        required_vars = ["SALT", "ENCRYPTION_KEY"]
        missing_vars = [var for var in required_vars if not os.getenv(var)]

        if missing_vars:
            raise SecurityError(
                f"Missing required environment variables: {', '.join(missing_vars)}"
            )

        salt_base64 = os.getenv("SALT", "")
        encryption_key_base64 = os.getenv("ENCRYPTION_KEY", "")

        return salt_base64, encryption_key_base64

    def _decrypt_master_key(self, encrypted_master_key: bytes) -> bytes:
        """Decrypt the master key."""
        salt_base64, encryption_key_base64 = self.load_and_validate_env()
        salt = base64.b64decode(salt_base64)
        encryption_key = base64.b64decode(encryption_key_base64)

        master_key = self.decrypt_master_key(encrypted_master_key, salt, encryption_key)

        self._secure_cleanup([encryption_key])
        return master_key

    def _decrypt_private_key(self, encrypted_private_key: bytes, master_key: bytes) -> PrivateKey:
        """Decrypt the private key using the master key."""
        box = SecretBox(master_key)
        private_key_bytes = box.decrypt(encrypted_private_key)
        private_key = PrivateKey(private_key_bytes)
        self._secure_cleanup([private_key_bytes])
        return private_key

    def _validate_keypair(self, private_key: PrivateKey, public_key: PublicKey) -> None:
        """Validate the keypair by performing a test encryption/decryption."""
        test_data = b"test"
        sealed_box = SealedBox(public_key)
        sealed_box_priv = SealedBox(private_key)

        encrypted = sealed_box.encrypt(test_data)
        decrypted = sealed_box_priv.decrypt(encrypted)

        if decrypted != test_data:
            raise SecurityError("Key pair validation failed")

    def __ensure_secure_folder(self) -> None:
        """Ensure the key folder exists with proper permissions."""
        if not os.path.exists(self.config["key_folder"]):
            os.makedirs(self.config["key_folder"], mode=0o700)
        else:
            os.chmod(self.config["key_folder"], 0o700)


class AccountManager:
    MAX_QUOTA = 16
    ACCOUNT_FILE_PATH = os.path.join(ConfigManager.get_system_config_dir(), "accounts.yaml")

    def __init__(self, encryptor, logger):
        self.encryptor = encryptor
        self.lock = threading.RLock()
        self.accounts = self.load_yaml()
        self.check_accounts()
        self.logger = logger
        atexit.register(self.save_yaml)

    def check_accounts(self):
        """檢查所有帳號的 exceed_time 是否超過 24 小時，若超過則清除 exceed_time 並將重置 exceed_quota."""
        now = datetime.now()
        update = False

        for _, account in self.accounts.items():
            exceed_time = account.get("exceed_time")
            if exceed_time and exceed_time != "Null":
                exceed_time_time = datetime.strptime(exceed_time, "%Y-%m-%dT%H:%M:%S")
                if now - exceed_time_time > timedelta(hours=24):
                    account["exceed_time"] = "Null"
                    account["exceed_quota"] = False
                    update = True

        if update:
            self.save_yaml()

    def create_account(self, username: str, password: str, public_key: PublicKey):
        with self.lock:
            encrypted_password = self.encryptor.encrypt_password(password, public_key)
            self.accounts[username] = {
                "encrypted_password": encrypted_password,
                "created_at": datetime.now().strftime("%Y-%m-%dT%H:%M:%S"),
                "exceed_quota": False,
                "exceed_time": "Null",
            }
        self.logger.info("Account %s has been created.", username)
        self.save_yaml()

    def verify_password(self, username: str, password: str, private_key: PrivateKey) -> bool:
        account = self.accounts.get(username)
        if not account:
            self.logger.error("Account does not exist.")
            return False

        encrypted_password = account.get("encrypted_password")
        decrypted_password = self.encryptor.decrypt_password(encrypted_password, private_key)
        if decrypted_password == password:
            print("Password is correct.")
            return True
        else:
            print("Incorrect password.")
            return False

    def read_account(self, username: str) -> dict:
        return self.accounts.get(username, {})

    def edit_account(
        self,
        public_key: PublicKey,
        old_username: str,
        new_username: str = "",
        new_password: str = "",
    ):
        with self.lock:
            if old_username in self.accounts:
                if new_username:
                    self.accounts[new_username] = self.accounts.pop(old_username)
                if new_password:
                    encrypted_password = self.encryptor.encrypt_password(new_password, public_key)
                    self.accounts[new_username or old_username]["encrypted_password"] = (
                        encrypted_password
                    )
                self.logger.info("Account %s has been updated.", old_username)
            else:
                self.logger.error("Account not found.")

    def delete_account(self, username: str):
        with self.lock:
            if username in self.accounts:
                del self.accounts[username]
                self.logger.info("Account %s has been deleted.", username)
            else:
                self.logger.error("Account %s not found.", username)
            self.save_yaml()

    def save_yaml(self):
        with self.lock:
            with open(self.ACCOUNT_FILE_PATH, "w") as file:
                yaml.dump(self.accounts, file, default_flow_style=False)
        # self.logger.info("Successfully update accounts information.")

    def load_yaml(self) -> dict:
        try:
            with open(self.ACCOUNT_FILE_PATH) as file:
                return yaml.safe_load(file) or {}
        except FileNotFoundError:
            return {}

    def get_account(self, private_key: PrivateKey) -> tuple[str, str]:
        eligible_accounts = {k: v for k, v in self.accounts.items() if not v["exceed_quota"]}

        if not eligible_accounts:
            self.logger.error("All accounts have exhausted their reading quota.")
            sys.exit(1)

        username, account = random.choice(list(eligible_accounts.items()))
        enc_pw = account["encrypted_password"]
        dec_pw = self.encryptor.decrypt_password(enc_pw, private_key)

        return username, dec_pw

    def update(self, username: str, field: str, new_value):
        with self.lock:
            account = self.accounts.get(username)
            if account:
                if field in account:
                    account[field] = new_value
                    self.logger.info("Updated %s for account %s.", field, username)
                    self.save_yaml()
                else:
                    self.logger.error("Field '%s' does not exist in the account.", field)
            else:
                self.logger.error("Account %s not found.", username)


class SecureFileHandler:
    @staticmethod
    def write_secure_file(path: str, data: bytes, permissions: int = 0o400) -> None:
        """Securely write data to a file with specified permissions."""
        with open(path, "wb") as f:
            f.write(data)
        os.chmod(path, permissions)

    @staticmethod
    def read_secure_file(path: str) -> bytes:
        """Securely read data from a file."""
        if not os.path.exists(path):
            raise SecurityError(f"Required file not found: {path}")
        with open(path, "rb") as f:
            return f.read()



# cli interface of _password.py
import getpass
import logging
import os
import sys
from collections import OrderedDict
from typing import Any

import questionary
from nacl.public import PrivateKey, PublicKey

from .utils.security_utils import AccountManager, Encryptor


def display_menu() -> Any:
    choices = [
        {"name": "Create Account", "value": "create"},
        {"name": "Read Account", "value": "read"},
        {"name": "Update Account", "value": "update"},
        {"name": "Delete Account", "value": "delete"},
        {"name": "Password Test", "value": "password"},
        {"name": "List All Accounts", "value": "list"},
        {"name": "Quit", "value": "quit"},
    ]

    answer = questionary.select("Please choose an action:", choices=choices).ask()
    return answer


def clean_terminal() -> None:
    os.system("cls" if os.name == "nt" else "clear")


def create_account(am: AccountManager, public_key: PublicKey) -> None:
    clean_terminal()
    print("Create Account")
    username = input("Please enter the username: ")
    if username == "":
        return
    password = getpass.getpass("Please enter the password: ")
    am.create_account(username, password, public_key)


def read_account(am: AccountManager) -> None:
    clean_terminal()
    print("Read Account")
    username = input("Please enter the username: ")
    account = am.read_account(username)
    if account:
        ordered_dict = OrderedDict()
        ordered_dict["username"] = username
        ordered_dict["encrypted_password"] = account["encrypted_password"]
        for key, value in account.items():
            if key not in ["username", "encrypted_password"]:
                ordered_dict[key] = value

        # Output OrderedDict
        for key, value in ordered_dict.items():
            print(f"{key}: {value}")
    else:
        print("Account not found.")


def update_account(am: AccountManager, public_key: PublicKey, private_key: PrivateKey) -> None:
    clean_terminal()
    print("Update Account")
    old_username = input("Please enter the old username: ")
    password = getpass.getpass("Please enter the password: ")
    if not am.verify_password(old_username, password, private_key):
        return
    new_username = input("Please enter the new username (leave blank to not update): ")
    password = getpass.getpass("Please enter the new password (leave blank to not update): ")
    am.edit_account(public_key, old_username, new_username, password)


def delete_account(am: AccountManager, private_key: PrivateKey) -> None:
    clean_terminal()
    print("Delete Account")
    username = input("Please enter the username: ")
    if username in am.accounts:
        password = getpass.getpass("Please enter the password: ")
        if not am.verify_password(username, password, private_key):
            return

        confirm_delete = questionary.select(
            f"Are you sure you want to delete the account {username}?",
            choices=[
                "Confirm",
                "Cancel",
            ],
        ).ask()

        if confirm_delete == "Confirm":
            am.delete_account(username)
        else:
            print("Operation canceled.")
    else:
        print("Account not found.")


def password_test(am: AccountManager, private_key: PrivateKey) -> None:
    clean_terminal()
    print("Password Test")
    username = input("Please enter the username: ")
    account = am.read_account(username)
    if account:
        password = getpass.getpass("Enter password: ")
        am.verify_password(username, password, private_key)
    else:
        print("Account not found.")


def list_accounts(am: AccountManager) -> None:
    clean_terminal()
    print("Account List")
    accounts = am.accounts
    if accounts:
        for username, info in accounts.items():
            print(
                f"Account: {username}, Exceed quota: {info['exceed_quota'] or 'Null'}, Exceed time: {info['exceed_time'] or 'Null'}, Created At: {info['created_at']}"
            )
    else:
        print("No accounts available.")


def cli() -> None:
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    clean_terminal()
    encryptor = Encryptor(logger)
    am = AccountManager(encryptor, logger)
    private_key, public_key = encryptor.load_keys()

    def execute_action(choice: str) -> bool:
        if choice == "create":
            create_account(am, public_key)
        elif choice == "read":
            read_account(am)
        elif choice == "update":
            update_account(am, public_key, private_key)
        elif choice == "delete":
            delete_account(am, private_key)
        elif choice == "password":
            password_test(am, private_key)
        elif choice == "list":
            list_accounts(am)
        elif choice == "quit":
            clean_terminal()
            print("Exiting the account management of v2dl.")
            return True
        else:
            print("Invalid choice, please try again.")
        return False

    while True:
        choice = display_menu()
        if execute_action(choice):
            break

    sys.exit(0)


if __name__ == "__main__":
    cli()



import random
import subprocess
import sys
import time
from subprocess import Popen

from selenium import webdriver
from selenium.common.exceptions import NoSuchElementException, TimeoutException
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.support import expected_conditions as EC  # noqa: N812
from selenium.webdriver.support.ui import WebDriverWait

from ..const import SELENIUM_AGENT
from .base import BaseBehavior, BaseBot, BaseScroll

DEFAULT_BOT_OPT = [
    "--remote-debugging-port=9222",
    "--disable-gpu",
    "--disable-infobars",
    "--disable-extensions",
    "--disable-dev-shm-usage",
    # "--disable-blink-features",
    # "--disable-blink-features=AutomationControlled",
    "--start-maximized",
    # "--ignore-certificate-errors",
]


class SeleniumBot(BaseBot):
    def __init__(self, runtime_config, base_config):
        super().__init__(runtime_config, base_config)
        self.init_driver()
        self.scroller = SelScroll(self.driver, self.config, self.logger)
        self.cloudflare = SelCloudflareHandler(self.driver, self.logger)

    def init_driver(self):
        self.driver: WebDriver
        self.chrome_process: Popen
        options = Options()

        chrome_path = [self.config.chrome.exec_path]
        # "or" returns the first True element
        subprocess_cmd = chrome_path + (self.runtime_config.chrome_args or DEFAULT_BOT_OPT)
        subprocess_cmd = [*subprocess_cmd, self.runtime_config.user_agent or SELENIUM_AGENT]

        if not self.runtime_config.use_chrome_default_profile:
            user_data_dir = self.prepare_chrome_profile()
            subprocess_cmd.append(f"--user-data-dir={user_data_dir}")

        self.chrome_process = subprocess.Popen(subprocess_cmd)

        options.add_experimental_option("debuggerAddress", "127.0.0.1:9222")
        try:
            self.driver = webdriver.Chrome(service=Service(), options=options)
        except Exception as e:
            self.logger.error("Unable to start Selenium WebDriver: %s", e)
            sys.exit("Unable to start Selenium WebDriver")

        # width = random.randint(1024, 1920)
        # height = random.randint(768, 1080)
        # self.driver.set_window_size(1920, 1080)

    def close_driver(self):
        pass
        # if self.close_browser:
        #     self.driver.quit()
        #     self.chrome_process.terminate()

    def auto_page_scroll(
        self, url: str, max_retry: int = 3, page_sleep: int = 5, fast_scroll: bool = False
    ) -> str:
        """Scroll page automatically with retries and Cloudflare challenge handle.

        The main function of this class.

        Args:
            url (str): Target URL
            max_retry (int): Maximum number of retry attempts. Defaults to 3
            page_sleep (int): The sleep time after reaching page bottom
            fast_scroll (bool): Whether to use fast scroll. Might be blocked by Cloudflare

        Returns:
            str: Page HTML content or error message
        """
        response: str = ""
        for attempt in range(max_retry):
            try:
                self.driver.get(url)
                SelBehavior.random_sleep(0.1, 0.5)

                if not self.handle_redirection_fail(url, max_retry, page_sleep):
                    self.logger.error(
                        "Reconnection fail for URL %s. Please check your network status.", url
                    )
                    break

                if self.cloudflare.handle_simple_block(attempt, max_retry):
                    continue

                # WebDriverWait(self.driver, 5).until(
                #     EC.presence_of_element_located((By.CSS_SELECTOR, "div.album-photo.my-2"))
                # )

                # main business
                self.handle_login()
                self.driver.execute_script("document.body.style.zoom='75%'")
                self.scroller.scroll_to_bottom()
                SelBehavior.random_sleep(5, 15)

                response = self.driver.page_source
                break

            except Exception as e:
                self.logger.exception(
                    "Request failed for URL %s - Attempt %d/%d. Error: %s",
                    url,
                    attempt + 1,
                    max_retry,
                    e,
                )

            self.logger.debug("Scrolling finished, pausing to avoid blocking")
            SelBehavior.random_sleep(page_sleep, page_sleep + 5)

        if not response:
            error_template = "Failed to retrieve URL after {} attempts: '{}'"
            error_msg = error_template.format(max_retry, url)
            self.logger.error(error_msg)
            return error_msg
        return response

    def handle_redirection_fail(self, url: str, max_retry: int, sleep_time: int) -> bool:
        if self.driver.current_url == url:
            return True
        retry = 1
        while retry <= max_retry:
            self.logger.error("Connection failed - Attempt %d/%d", retry, max_retry)
            SelBehavior.random_sleep(sleep_time, sleep_time + 5 * random.uniform(1, retry * 5))

            if self.cloudflare.handle_simple_block(retry, max_retry):
                self.logger.critical("Failed to solve Cloudflare turnstile challenge")
                continue

            self.driver.get(url)
            retry += 1
            if self.driver.current_url == url:
                return True

        return self.driver.current_url == url

    def handle_login(self):
        success = False
        if self.driver.find_elements(
            By.XPATH, "//h1[@class='h4 text-secondary mb-4 login-box-msg']"
        ):
            self.logger.info("Login page detected - Starting login process")
            try:
                self.email, self.password = self.account_manager.get_account(self.private_key)
                if self.email is None or self.password is None:
                    self.logger.critical("Email and password not provided")
                    sys.exit("Automated login failed.")

                email_field = self.driver.find_element(By.ID, "email")
                password_field = self.driver.find_element(By.ID, "password")
                BaseBehavior.random_sleep(0.5, 1)

                # SelBehavior.human_like_mouse_movement(self.driver, email_field)
                password_field.send_keys(Keys.SHIFT, Keys.TAB)
                BaseBehavior.random_sleep(0.5, 1)
                SelBehavior.human_like_type(email_field, self.email)
                SelBehavior.random_sleep(0.01, 0.3)

                # SelBehavior.human_like_mouse_movement(self.driver, password_field)
                # SelBehavior.human_like_click(self.driver, email_field)
                email_field.send_keys(Keys.TAB)
                SelBehavior.human_like_type(password_field, self.password)
                SelBehavior.random_sleep(0.01, 0.5)

                # try:
                #     remember_checkbox = self.driver.find_element(By.ID, "remember")
                #     if not remember_checkbox.is_selected():
                #         SelBehavior.human_like_click(self.driver, remember_checkbox)
                # except NoSuchElementException:
                #     self.logger.warning("Remember me checkbox not found")

                try:
                    self.cloudflare.handle_cloudflare_recaptcha()
                except Exception as e:
                    self.logger.exception("Error handling Cloudflare reCAPTCHA: %s", e)

                self.driver.find_element(
                    By.XPATH, '//button[@type="submit" and @class="btn btn-primary btn-block"]'
                ).click()

                SelBehavior.random_sleep(3, 5)

                if not self.driver.find_elements(
                    By.XPATH, "//h1[@class='h4 text-secondary mb-4 login-box-msg']"
                ):
                    self.logger.info("Login successful")
                    success = True
                else:
                    self.logger.error("Login failed - Checking for error messages")
                    self.check_login_errors()

            except NoSuchElementException as e:
                self.logger.error("Login form element not found: %s", e)
            except TimeoutException as e:
                self.logger.error("Timeout waiting for element: %s", e)
            except Exception as e:
                self.logger.error("Unexpected error during login: %s", e)
        else:
            success = True
        if not success:
            self.logger.critical("Automated login failed. Please login yourself.")
            sys.exit("Automated login failed.")

    def check_login_errors(self):
        error_messages = self.driver.find_elements(By.CLASS_NAME, "errorMessage")
        if error_messages:
            for message in error_messages:
                self.logger.error("Login error: %s", message.text)
        else:
            self.logger.warning(
                "No specific error message found. Login might have failed for unknown reasons."
            )


class SelCloudflareHandler:
    """Handles Cloudflare protection detection and bypass attempts.

    Includes methods for dealing with various Cloudflare challenges.
    """

    def __init__(self, driver: WebDriver, logger):
        self.driver = driver
        self.logger = logger

    def handle_simple_block(self, attempt: int, retries: int) -> bool:
        """check and handle Cloudflare challenge."""
        blocked = False
        if self.is_simple_blocked():
            self.logger.info(
                "Detected Cloudflare challenge, attempting to solve... Attempt %d/%d",
                attempt + 1,
                retries,
            )
            self.handle_cloudflare_turnstile()
            blocked = True
        return blocked

    def handle_hard_block(self) -> bool:
        """Check, log critical, and return whether blocked or not.

        This is a cloudflare WAF full page block.
        """
        blocked = False
        if self.is_hard_block():
            self.logger.critical("Hard block detected by Cloudflare - Unable to proceed")
            blocked = True
        return blocked

    def is_simple_blocked(self) -> bool:
        title_check = any(text in self.driver.title for text in ["請稍候...", "Just a moment..."])
        page_source_check = "Checking your" in self.driver.page_source
        return title_check or page_source_check

    def is_hard_block(self) -> bool:
        is_blocked = "Attention Required! | Cloudflare" in self.driver.title
        if is_blocked:
            self.logger.critical("Cloudflare hard block detected")
        return is_blocked

    def handle_cloudflare_turnstile(self):
        try:
            iframe = WebDriverWait(self.driver, 10).until(
                EC.presence_of_element_located((
                    By.XPATH,
                    "//iframe[contains(@src, 'challenges.cloudflare.com')]",
                ))
            )
            self.driver.switch_to.frame(iframe)

            checkbox = WebDriverWait(self.driver, 10).until(
                EC.element_to_be_clickable((By.ID, "cf-turnstile-response"))
            )
            SelBehavior.human_like_click(self.driver, checkbox)

            if "Select all squares with" in self.driver.page_source:
                self.solve_image_captcha()

            self.driver.switch_to.default_content()
            SelBehavior.random_sleep(10, 20)
        except (TimeoutException, NoSuchElementException):
            self.logger.error("Unable to solve Cloudflare challenge.")

    def handle_cloudflare_recaptcha(self):
        try:
            recaptcha_checkbox = WebDriverWait(self.driver, 10).until(
                EC.element_to_be_clickable((By.XPATH, "//input[@type='checkbox']"))
            )
            SelBehavior.human_like_click(self.driver, recaptcha_checkbox)

            verify_button = WebDriverWait(self.driver, 10).until(
                EC.element_to_be_clickable((By.XPATH, "//span[contains(text(), '驗證您是人類')]"))
            )
            SelBehavior.human_like_click(self.driver, verify_button)

            SelBehavior.random_sleep(3, 5)
        except (TimeoutException, NoSuchElementException) as e:
            self.logger.warning(
                "reCAPTCHA checkbox or verify button not found or unable to interact: %s", e
            )

    def solve_image_captcha(self):
        raise NotImplementedError


class SelBehavior(BaseBehavior):
    @staticmethod
    def human_like_mouse_movement(driver, element):
        action = ActionChains(driver)
        action.move_by_offset(random.randint(-100, 100), random.randint(-100, 100))
        action.move_to_element_with_offset(
            element, random.randint(-10, 10), random.randint(-10, 10)
        )
        action.pause(random.uniform(0.1, 0.3))
        action.move_to_element(element)
        action.perform()
        SelBehavior.random_sleep(*BaseBehavior.pause_time)

    @staticmethod
    def human_like_click(driver, element):
        SelBehavior.human_like_mouse_movement(driver, element)
        action = ActionChains(driver)
        action.click()
        action.perform()
        SelBehavior.random_sleep(*BaseBehavior.pause_time)

    @staticmethod
    def human_like_type(element, text):
        for char in text:
            element.send_keys(char)
            time.sleep(random.uniform(0.001, 0.2))
        SelBehavior.random_sleep(*BaseBehavior.pause_time)


class SelScroll(BaseScroll):
    def __init__(self, driver: WebDriver, config, logger):
        super().__init__(config, logger)
        self.driver = driver

    def scroll_to_bottom(self):
        max_attempts = 10
        attempts = 0
        last_position = -123459
        scroll_length = lambda: random.randint(
            self.config.download.min_scroll_length,
            self.config.download.max_scroll_length,
        )

        while attempts < max_attempts:
            SelBehavior.random_sleep(1, 2)
            scroll = scroll_length()
            self.logger.debug(
                "Current position: %d, scrolling down by %d pixels", last_position, scroll
            )
            self.driver.execute_script(f"window.scrollBy({{top: {scroll}, behavior: 'smooth'}});")
            new_position = self.driver.execute_script("return window.pageYOffset;")
            if new_position == last_position:
                break
            last_position = new_position
            attempts += 1

    def old_scroll_to_bottom(self):
        self.logger.info("Start scrolling the page")
        scroll_attempts = 0
        max_attempts = 45

        scroll_pos_init = self.driver.execute_script("return window.pageYOffset;")
        step_scroll = random.randint(
            self.config.download.min_scroll_length,
            self.config.download.max_scroll_length,
        )

        while scroll_attempts < max_attempts:
            scroll_attempts += 1

            self.driver.execute_script(f"window.scrollBy(0, {step_scroll});")
            scroll_pos_end = self.driver.execute_script("return window.pageYOffset;")
            time.sleep(0.75)

            if scroll_pos_init >= scroll_pos_end:
                self.logger.debug("Reached the bottom of the page")
                break

            scroll_pos_init = scroll_pos_end

            step_scroll = random.randint(
                self.config.download.min_scroll_length, self.config.download.max_scroll_length
            )

            self.wait_for_content_load()

            self.successive_scroll_count += 1
            if self.successive_scroll_count >= self.max_successive_scrolls:
                pause_time = random.uniform(3, 7)
                self.logger.debug(
                    "Scrolled %d times, pausing for %.2f seconds",
                    self.successive_scroll_count,
                    pause_time,
                )
                time.sleep(pause_time)
                self.successive_scroll_count = 0
                self.max_successive_scrolls = random.randint(3, 7)

        if scroll_attempts == max_attempts:
            self.logger.info(
                "Reached maximum attempts (%d), scrolling finished, may not have fully reached the bottom",
                max_attempts,
            )
        else:
            self.logger.info("Page scroll completed")

    def perform_scroll_action(self):
        action = random.choices(
            ["scroll_down", "scroll_up", "pause", "jump"],
            weights=[0.7, 0.1, 0.1, 0.1],
        )[0]

        current_position = self.get_scroll_position()

        if action == "scroll_down":
            scroll_length = random.randint(
                self.config.download.min_scroll_length,
                self.config.download.max_scroll_length,
            )
            target_position = current_position + scroll_length
            self.logger.debug("Trying to scroll down %d pixels", scroll_length)
            actual_position = self.safe_scroll(target_position)
            self.logger.debug("Actually scrolled to %d pixels", actual_position)
            time.sleep(random.uniform(*BaseBehavior.pause_time))
        elif action == "scroll_up":
            scroll_length = random.randint(
                self.config.download.min_scroll_length,
                self.config.download.max_scroll_length,
            )
            target_position = max(0, current_position - scroll_length)
            self.logger.debug("Trying to scroll up %d pixels", scroll_length)
            actual_position = self.safe_scroll(target_position)
            self.logger.debug("Actually scrolled to %d pixels", actual_position)
        elif action == "pause":
            pause_time = random.uniform(1, 3)
            self.logger.debug("Pausing for %.2f seconds", pause_time)
            time.sleep(pause_time)
        elif action == "jump":
            jump_position = current_position + random.randint(100, 500)
            self.logger.debug("Trying to jump to position %d", jump_position)
            actual_position = self.safe_scroll(jump_position)
            self.logger.debug("Actually jumped to %d", actual_position)

    def safe_scroll(self, target_position):
        current_position = self.get_scroll_position()
        step = random.uniform(
            self.config.download.min_scroll_step,
            self.config.download.max_scroll_step,
        )
        # step = 50 if target_position > current_position else -50
        # while abs(current_position - target_position) > abs(step):

        while abs(current_position - target_position) > step:
            self.driver.execute_script(f"window.scrollTo(0, {current_position + step});")
            time.sleep(random.uniform(0.01, 0.2))
            new_position = self.get_scroll_position()
            if new_position == current_position:
                self.logger.debug(
                    "Cannot continue scrolling, target: %d, current: %d",
                    target_position,
                    current_position,
                )
                break
            current_position = new_position
        self.driver.execute_script(f"window.scrollTo(0, {target_position});")
        return self.get_scroll_position()

    def get_scroll_position(self):
        return self.driver.execute_script("return window.pageYOffset")

    def get_page_height(self):
        return self.driver.execute_script("return document.body.scrollHeight")

    def wait_for_content_load(self):
        try:
            WebDriverWait(self.driver, 5).until(
                lambda d: d.execute_script("return document.readyState") == "complete"
            )
        except TimeoutException:
            self.logger.warning("Timeout waiting for new content to load")



# v2dl/utils/__init__.py
from .security_utils import AccountManager, Encryptor, SecureFileHandler
from .utils import AlbumTracker, LinkParser, ThreadingService, ThreadJob

# only import __all__ when using from automation import *
__all__ = [
    "AccountManager",
    "Encryptor",
    "SecureFileHandler",
    "AlbumTracker",
    "LinkParser",
    "ThreadJob",
    "ThreadingService",
]
